# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = "libf3d"
copyright = "2025, Michael Migliore and Mathieu Westphal"
author = "Michael Migliore and Mathieu Westphal"

# The full version, including alpha/beta/rc tags
release = "3.2.0"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ["breathe", "exhale"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_book_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Setup the breathe extension
breathe_projects = {"libf3d": "./_doxygen/xml"}
breathe_default_project = "libf3d"

# Setup the exhale extension
exhale_args = {
    # These arguments are required
    "containmentFolder": "./api",
    "rootFileName": "library_root.rst",
    "doxygenStripFromPath": "..",
    # Heavily encouraged optional argument (see docs)
    "rootFileTitle": "Library API",
    # Suggested optional arguments
    "createTreeView": True,
    # TIP: if using the sphinx-bootstrap-theme, you need
    # "treeViewIsBootstrap": True,
    "exhaleExecutesDoxygen": True,
    "exhaleDoxygenStdin": "INPUT = ../public\nDISTRIBUTE_GROUP_DOC = YES\nEXCLUDE_SYMBOLS = std",
}

# Tell sphinx what the primary language being documented is.
primary_domain = "cpp"

# Tell sphinx what the pygments highlight language should be.
highlight_language = "cpp"



from pathlib import Path

import pytest

import f3d


@pytest.fixture
def test_plugins_list():
    base_dir = Path(f3d.__file__).parent
    plugins = f3d.Engine.get_plugins_list(base_dir / "share/f3d/plugins")
    plugins += f3d.Engine.get_plugins_list(base_dir / "../share/f3d/plugins")
    plugins += f3d.Engine.get_plugins_list(base_dir / "../../share/f3d/plugins")

    assert len(plugins) > 0
    assert plugins.index("native") >= 0


def test_get_lib_info():
    lib_info = f3d.Engine.get_lib_info()

    assert isinstance(lib_info, f3d.LibInformation)

    assert isinstance(lib_info.version, str) and lib_info.version
    assert isinstance(lib_info.version_full, str) and lib_info.version_full
    assert isinstance(lib_info.build_date, str) and lib_info.build_date
    assert isinstance(lib_info.build_system, str) and lib_info.build_system
    assert isinstance(lib_info.compiler, str) and lib_info.compiler

    assert isinstance(lib_info.modules, dict) and lib_info.modules
    for key, value in lib_info.modules.items():
        assert isinstance(key, str) and key
        assert isinstance(value, bool)

    assert isinstance(lib_info.vtk_version, str) and lib_info.vtk_version

    assert isinstance(lib_info.copyrights, list) and lib_info.copyrights
    for copyright in lib_info.copyrights:
        assert isinstance(copyright, str) and copyright

    assert isinstance(lib_info.license, str) and lib_info.license


def test_get_readers_info():
    readers_info = f3d.Engine.get_readers_info()

    assert isinstance(readers_info, list) and readers_info

    for reader in readers_info:
        assert isinstance(reader, f3d.ReaderInformation)

        assert isinstance(reader.name, str) and reader.name
        assert isinstance(reader.description, str) and reader.description

        assert isinstance(reader.extensions, list) and reader.extensions
        for ext in reader.extensions:
            assert isinstance(ext, str) and ext

        assert isinstance(reader.mime_types, list) and reader.mime_types
        for mime in reader.mime_types:
            assert isinstance(mime, str) and mime

        assert isinstance(reader.plugin_name, str) and reader.plugin_name
        assert isinstance(reader.has_scene_reader, bool)
        assert isinstance(reader.has_geometry_reader, bool)


def test_get_rendering_backend_list():
    backends = f3d.Engine.get_rendering_backend_list()

    assert isinstance(backends, dict) and len(backends) == 5


def test_reader_options():
    reader_option_names = f3d.Engine.get_all_reader_option_names()
    assert reader_option_names is not None

    with pytest.raises(KeyError):
        f3d.Engine.set_reader_option("inexistent", "value")



from argparse import ArgumentParser
from pathlib import Path

import f3d

TEST_DATA_DIR = Path(__file__).parent.parent.parent.parent.parent / "testing/data"

if __name__ == "__main__":
    argparser = ArgumentParser()
    argparser.add_argument("models", nargs="*")

    args = argparser.parse_args()

    input_paths = args.models or [TEST_DATA_DIR / "suzanne.obj"]

    # Create a native window engine
    eng = f3d.Engine.create()

    # Add a model
    try:
        eng.scene.add(input_paths)
    except RuntimeError as e:
        print(e)

    # Render
    eng.window.render()

    eng.interactor.start()



import sys
from argparse import ArgumentParser
from pathlib import Path
from tempfile import gettempdir

import f3d

TEST_DATA_DIR = Path(__file__).parent.parent.parent.parent.parent / "testing/data"

if __name__ == "__main__":
    argparser = ArgumentParser()
    argparser.add_argument("models", nargs="*")
    argparser.add_argument("--into", default=gettempdir())

    args = argparser.parse_args()

    input_paths = args.models or [TEST_DATA_DIR / "suzanne.obj"]
    output_path = Path(args.into) / "f3d.png"

    try:
        eng = f3d.Engine.create(True)
        eng.window.size = 300, 300

        eng.scene.add(input_paths)

        img = eng.window.render_to_image()
        img.save(output_path)

        print(f"rendered to {output_path}")
    except Exception as e:
        print("F3D encountered an unexpected exception:")
        print(e)
        sys.exit(1)



import tkinter as tk  # Works also on CustomTKinter

import f3d
from pyopengltk import OpenGLFrame


class Frame(OpenGLFrame):
    # !!!OpenGlFrame requires adding own code to initgl and redraw!!!
    # !!!This solution only renders the F3D viewer, as it uses EXTERNAL, so controls need to be defined!!!
    def __init__(self):
        super().__init__()
        self.engine = None

    # Initialize F3D
    def initgl(self):
        self.engine = (
            f3d.Engine.create_external_glx()
        )  # use create_external_egl for Wayland
        self.engine.scene.add(
            f3d.Mesh(
                points=[0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0],
                face_sides=[3],
                face_indices=[0, 1, 2],
            )
        )

    def redraw(self):
        self.engine.window.render()


if __name__ == "__main__":
    # Create main window and define size, position and title
    root = tk.Tk()
    root.geometry("640x480+100+100")
    root.title("Minimal tkinter interface")
    # Create OpenGL instance
    F3D = Frame()
    F3D.pack(fill=tk.BOTH, expand=tk.YES)

    # Run TKinter mainloop
    root.mainloop()



import math
import os
import time
from pathlib import Path
from typing import Any

import f3d

TEST_DATA_DIR = Path(__file__).parent.parent.parent.parent.parent / "testing/data"


def main():
    model_path = TEST_DATA_DIR / "suzanne.obj"
    options: dict[str, Any] = {
        "scene.up_direction": "+Y",
        "render.effect.tone_mapping": True,
        "render.effect.ambient_occlusion": True,
        "render.effect.translucency_support": True,
        "render.effect.antialiasing.enable": True,
    }
    anim_fps = 30
    anim_duration = 4
    orbit_count = 2

    try:
        rows, cols = os.get_terminal_size()
    except OSError:
        rows, cols = 40, 20

    # setup offscreen engine
    engine = f3d.Engine.create(offscreen=True)
    engine.options.update(options)
    engine.scene.add(model_path)
    engine.window.size = rows, cols * 2

    # fit view to loaded model and grab computed camera position
    engine.window.camera.reset_to_bounds()
    x0, y0, z0 = engine.window.camera.position
    r = math.sqrt(x0**2 + z0**2)

    # loop to move the camera and render frames
    for t in realtime_animation_timesteps(anim_fps, anim_duration):
        # set camera using polar coordinates on XZ plane, and original Y altitude
        a = (t / anim_duration * orbit_count) * 2 * math.pi
        engine.window.camera.position = math.sin(a) * r, y0, math.cos(a) * r
        engine.window.camera.view_up = 0, 1, 0

        # render
        image = engine.window.render_to_image(no_background=True)
        print(image.to_terminal_text(), end="")
    print()


def realtime_animation_timesteps(fps: float, duration: float):
    """Generate times at `fps`frame/s accounting for actual frame computation time."""
    target_frame_duration = 1.0 / fps
    start_time = time.time()
    end_time = start_time + duration
    while time.time() < end_time:
        frame_start_time = time.time()

        yield frame_start_time - start_time

        actual_frame_duration = time.time() - frame_start_time
        time.sleep(max(target_frame_duration - actual_frame_duration, 0))


if __name__ == "__main__":
    main()



from typing import Any
import pytest

import f3d


def test_closest_option():
    option, distance = f3d.Options().get_closest_option("scene_direction")
    assert option == "scene.up_direction"
    assert distance == 3


def test_setitem():
    options = f3d.Options()
    options["ui.axis"] = False
    options["render.background.blur.coc"] = 33.33
    options["scene.animation.speed_factor"] = 3.3
    options["render.raytracing.samples"] = 5
    options["render.grid.color"] = [1.0, 1.0, 1.0]
    options["model.point_sprites.type"] = "sphere"


def test_getitem():
    engine = f3d.Engine.create_none()
    options = engine.options

    assert options["ui.axis"] is False
    assert options["render.background.blur.coc"] == 20.0
    assert options["scene.animation.speed_factor"] == 1.0
    assert options["render.raytracing.samples"] == 5
    assert options["render.grid.color"] == [0.0, 0.0, 0.0]
    assert options["model.point_sprites.type"] == "sphere"


def test_get_non_existent_key():
    options = f3d.Options()
    with pytest.raises(KeyError):
        assert options["hello"] == "world"


def test_set_non_existent_key():
    options = f3d.Options()
    with pytest.raises(KeyError):
        options["hello"] = "world"


def test_set_invalid_str_value():
    options = f3d.Options()
    with pytest.raises(ValueError):
        options["ui.axis"] = "world"


def test_set_incompatible_value_type():
    options = f3d.Options()
    with pytest.raises(AttributeError):
        options["ui.axis"] = 1.12


def test_len():
    options = f3d.Options()
    assert len(options) == len(options.keys())


def test_iter():
    options = f3d.Options()
    assert list(options) == options.keys()


def test_contains():
    options = f3d.Options()
    assert "render.grid.color" in options
    assert "hello.world" not in options


def test_set_options():
    options = f3d.Options()
    options["ui.axis"] = True
    options["model.material.roughness"] = 0.7
    options["scene.animation.speed_factor"] = 3.7
    options["render.raytracing.samples"] = 2
    options["model.color.rgb"] = [0.0, 1.0, 1.0]
    options["model.point_sprites.type"] = "other"

    engine = f3d.Engine.create_none()
    engine.options = options
    assert engine.options["ui.axis"] is True
    assert engine.options["model.material.roughness"] == 0.7
    assert engine.options["scene.animation.speed_factor"] == 3.7
    assert engine.options["render.raytracing.samples"] == 2
    assert engine.options["model.color.rgb"] == [0.0, 1.0, 1.0]
    assert engine.options["model.point_sprites.type"] == "other"


def test_set_options_from_string():
    options = f3d.Options()
    options["ui.axis"] = False
    assert not options["ui.axis"]

    options["ui.axis"] = "on"
    assert options["ui.axis"]

    options["ui.axis"] = "off"
    assert not options["ui.axis"]


def test_to_dict():
    options = f3d.Options()
    options_dict = dict(options)

    assert len(options) > 0
    assert len(options) == len(options_dict)
    for k, v in options_dict.items():
        assert options[k] == v


def test_update_from_dict():
    options = f3d.Options()

    d: dict[str, Any] = {
        "ui.axis": True,
        "model.material.roughness": 0.8,
        "scene.animation.speed_factor": 3.8,
        "render.raytracing.samples": 8,
        "model.color.rgb": [0.1, 0.2, 1.3],
        "model.point_sprites.type": "-X",
    }
    options.update(d)
    for k, v in d.items():
        assert options[k] == v, f"{k} was not set correctly"


def test_update_from_kv_pairs():
    options = f3d.Options()

    d: dict[str, Any] = {
        "ui.axis": True,
        "model.material.roughness": 0.8,
        "scene.animation.speed_factor": 3.8,
        "render.raytracing.samples": 8,
        "model.color.rgb": [0.1, 0.2, 1.3],
        "model.point_sprites.type": "-X",
    }
    options.update(d.items())
    for k, v in d.items():
        assert options[k] == v, f"{k} was not set correctly"


def test_update_from_invalid_kv_pairs():
    options = f3d.Options()

    items = (
        ("ui.axis", True),
        ("model.material.roughness", 0.8),
        ("a", "b", 3),
    )
    with pytest.raises(ValueError):
        options.update(items)  # type: ignore


def test_is_same():
    options1 = f3d.Options()
    options2 = f3d.Options()
    options1["ui.axis"] = True
    options2["ui.axis"] = False
    assert not options2.is_same(options1, "ui.axis")


def test_is_copy():
    options1 = f3d.Options()
    options2 = f3d.Options()
    options1["ui.axis"] = True
    options2["ui.axis"] = False
    assert not options2.is_same(options1, "ui.axis")
    options2.copy(options1, "ui.axis")
    assert options2.is_same(options1, "ui.axis")



import tempfile
from pathlib import Path

import f3d


def test_scene_memory():
    testing_dir = Path(__file__).parent.parent.parent / "testing"
    reference = Path(testing_dir) / "baselines/TestPythonSceneMemory.png"
    output = Path(tempfile.gettempdir()) / "TestPythonSceneMemory.png"

    engine = f3d.Engine.create(True)
    engine.window.size = 300, 300

    engine.scene.add(
        f3d.Mesh(
            points=[0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0],
            face_sides=[3],
            face_indices=[0, 1, 2],
        )
    )

    img = engine.window.render_to_image()
    img.save(output)

    assert img.compare(f3d.Image(reference)) < 0.05


def test_scene():
    testing_dir = Path(__file__).parent.parent.parent / "testing"
    world = testing_dir / "data/world.obj"
    logo = testing_dir / "data/f3d.glb"
    sphere1 = testing_dir / "data/mb/recursive/mb_1_0.vtp"
    sphere2 = testing_dir / "data/mb/recursive/mb_2_0.vtp"
    cube = testing_dir / "data/mb/recursive/mb_0_0.vtu"
    reference = testing_dir / "baselines/TestPythonScene.png"
    output = Path(tempfile.gettempdir()) / "TestPythonScene.png"

    engine = f3d.Engine.create(True)
    engine.window.size = 300, 300

    engine.scene.add([world, logo])
    engine.scene.add(sphere1)
    engine.scene.add([sphere2, cube])

    assert engine.scene.animation_time_range() == (0.0, 4.0)
    engine.scene.load_animation_time(2)

    assert engine.scene.available_animations() == 1

    img = engine.window.render_to_image()
    img.save(output)

    assert img.compare(f3d.Image(reference)) < 0.05



import f3d
import os

from pathlib import Path


def test_text_distance():
    assert f3d.Utils.text_distance("modle", "model") == 2


def test_collapse_path():
    assert f3d.Utils.collapse_path("/folder/../file.ext", ".") == Path("/file.ext")


def test_glob_to_regex():
    assert f3d.Utils.glob_to_regex("*vt?") == ".*vt."
    assert f3d.Utils.glob_to_regex("**/*vt?") == "(?:[^/]*(?:/|$))*[^/]*vt[^/]"
    assert (
        f3d.Utils.glob_to_regex(glob="**\\\\*vt?", path_separator="\\")
        == "(?:[^\\\\]*(?:\\\\|$))*[^\\\\]*vt[^\\\\]"
    )


def test_get_env():
    assert f3d.Utils.get_env("F3D_TEST_ENV_EMPTY") == None


def test_get_know_folder():
    if os.name == "nt":
        assert f3d.Utils.get_known_folder(f3d.Utils.KnownFolder.PICTURES) != None
    else:
        assert f3d.Utils.get_known_folder(f3d.Utils.KnownFolder.PICTURES) == None
